class CalendlyException(Exception):
    """Errors corresponding to a misuse of Calendly API"""

    def __init__(self, message=None, details=None):
        self.message = message or ""
        self.details = details or []
        super(CalendlyException, self).__init__(f"{self.message} - {self.details}")

class CalendlyOauth2Exception(CalendlyException):
    """Errors corresponding to a misuse of CalendlyOauth2 API"""


from typing import MutableMapping
from calendly.exceptions import CalendlyException
import requests

__author__ = "laxmena <ConnectWith@laxmena.com>"
__license__ = "MIT"


class CalendlyReq(object):
    """
    Private class wrapping the Calendly API v2. Decodes responses from Calendly and returns it

    References
    ----------
    https://calendly.stoplight.io/docs/api-docs/
    """

    OAUTH2_ERROR_TYPE_KEY = 'error'
    OAUTH2_ERROR_DESCRIPTION_KEY = 'error_description'

    API_ERROR_TYPE_KEY = "title"
    API_ERROR_DESCRIPTION_KEY = "message"
    API_ERROR_DETAILS_KEY = "details"

    def __init__(self, token: str=None, headers: dict=None):
        """
        Constructor: Uses Bearer Token Authentication or custom headers.

        Parameters
        ----------
        token : str 
        headers : str
            Personal Access Token
        """

        if token and headers:
            raise CalendlyException("You can't pass both token and headers at the same time.")

        if token:
            headers = {'authorization': 'Bearer ' + token}

        self.headers = headers

    def _get_oauth2_error_from_response(self,response):
        try:
            resp = response.json()
            return resp[self.OAUTH2_ERROR_TYPE_KEY], resp[self.OAUTH2_ERROR_DESCRIPTION_KEY], []
        except (AttributeError, KeyError):
            return

    def _get_api_error_from_response(self, response):

        try:
            resp = response.json()
            errors = [resp[self.API_ERROR_TYPE_KEY], resp[self.API_ERROR_DESCRIPTION_KEY]]
        except (AttributeError, KeyError):
            return

        try:
            errors.append(resp[self.API_ERROR_DETAILS_KEY])
        except (AttributeError, KeyError):
            errors.append([])

        return tuple(errors)

    def _get_error_type_and_description_from_response(self, response):

        oauth2_errors = self._get_oauth2_error_from_response(response)
        if not oauth2_errors:
            oauth2_errors = self._get_api_error_from_response(response)

        if not oauth2_errors:
            oauth2_errors = "error", "Unknown Error.", []

        return  oauth2_errors

    def process_request(self, method: str, url: str, data: MutableMapping=None) -> requests.Response:
        """
        Make requests to Calendly API by appending requried headers. 

        Parameters
        ----------
        method : str
            supported methods - get, post, delete, put
        url : str
            Calendly API URL
        data : dict, optional
            additional data to be passed to the API 
        """
        request_method = getattr(requests, method)
        kwargs = dict(json=data)

        if self.headers:
            kwargs.update(dict(headers=self.headers))

        response = request_method(url, **kwargs)
        if response.status_code > requests.codes.permanent_redirect:
            error_type, error_description, error_details = self._get_error_type_and_description_from_response(response)
            raise CalendlyException(f"{error_type}: {error_description}", error_details)

        return response

    def get(self, url: str, data: MutableMapping=None) -> requests.Response:
        """
        Send GET request to the Calendly URL.

        Parameters
        ----------
        url : str
            Calendly API URL
        data : dict, optional
            additional data to be passed to the API 
        """
        return self.process_request('get', url, data)

    def post(self, url: str, data: MutableMapping=None) -> requests.Response:
        """
        Send POST request to the Calendly URL.

        Parameters
        ----------
        url : str
            Calendly API URL
        data : dict, optional
            additional data to be passed to the API 
        """
        return self.process_request('post', url, data)

    def delete(self, url: str, data: MutableMapping=None) -> requests.Response:
        """
        Send DELETE request to the Calendly URL.

        Parameters
        ----------
        url : str
            Calendly API URL
        data : dict, optional
            additional data to be passed to the API 
        """
        return self.process_request('delete', url, data)

    def put(self, url: str, data: MutableMapping=None) -> requests.Response:
        """
        Send PUT request to the Calendly URL.

        Parameters
        ----------
        url : str
            Calendly API URL
        data : dict, optional
            additional data to be passed to the API 
        """        
        return self.process_request('put', url, data)


from .utils.oauth2 import CalendlyOauth2
from .calendly import CalendlyAPI
from .exceptions import CalendlyException, CalendlyOauth2Exception

__all__ = [CalendlyAPI, CalendlyOauth2, CalendlyException, CalendlyOauth2Exception]



BASE="https://api.calendly.com"
WEBHOOK=f"{BASE}/webhook_subscriptions"
USERS=f"{BASE}/users"
ME=f"{USERS}/me"
EVENT_TYPE=f"{BASE}/event_types"
EVENTS=f"{BASE}/scheduled_events"
WEBHOOK_SUBSCRIPTIONS=f"{BASE}/webhook_subscriptions"
SCHEDULING_LINKS=f"{BASE}/scheduling_links"
ORGANIZATIONS=f"{BASE}/organizations/"
ORGANIZATION_MEMBERSHIPS=f"{BASE}/organization_memberships"
DATA_COMPLIANCE=f"{BASE}/data_compliance/deletion/invitees"

OAUTH_BASE_URL = "https://auth.calendly.com/oauth"
OAUTH_AUTHORIZE_URL = f"{OAUTH_BASE_URL}/authorize"
OAUTH_TOKEN_URL = f"{OAUTH_BASE_URL}/token"
OAUTH_REVOKE_URL = f"{OAUTH_BASE_URL}/revoke"
OAUTH_INTROSPECT_URL = f"{OAUTH_BASE_URL}/introspect"




import json
from typing import List, MutableMapping

from calendly.utils.api import CalendlyReq
from calendly.utils.constants import WEBHOOK, EVENTS, ME, EVENT_TYPE
from calendly.exceptions import CalendlyException


class CalendlyAPI(object):

    event_types_def = {
        "canceled": "invitee.canceled",
        "created": "invitee.created"
    }

    def __init__(self, token: str):
        """
        Constructor. Uses Bearer Token for Authentication.

        Parameters
        ----------
        token : str 
            Personal Access Token
        """
        self.request = CalendlyReq(token)

    def create_webhook(self, url: str, scope: str, organization: str, signing_key: str=None, user: str=None, event_types: List[str]=("canceled", "created")) -> MutableMapping:
        """
        Create a Webhook Subscription

        Args:
            url (str): Webhook URL
            scope (str): Either "organization" or "user"
            organization (str): Unique reference to the organization that the webhook will be tied to
            signing_key (str): A signing key
            user (str, optional): If scope is set to "user", then user reference is required.
            event_types (list, optional): List of user events to subscribe to. Defaults to ["canceled", "created"].

        Returns:
            response: dict
                json-decoded response from Calendly API 
        """
        events = [self.event_types_def[event_type]
                  for event_type in event_types]
        data = {'url': url,
                'events': events,
                'organization': organization,
                'scope': scope,
                'signing_key': signing_key}

        if scope == 'user':
            if user is None:
                raise CalendlyException
            data['user'] = user

        response = self.request.post(WEBHOOK, data)
        return response.json()

    def list_webhooks(self, organization: str, scope: str, user: str=None, count: int=20, sort: str=None) -> List[MutableMapping]:
        """ 
        Get a List of Webhook subscriptions for an Organization or User with a UUID.
        Reference:
        https://calendly.stoplight.io/docs/api-docs/reference/calendly-api/openapi.yaml/paths/~1webhook_subscriptions/get

        Args:
            organization (str): Unique reference to the organization that the webhook will be tied to
            scope (str): Either "organization" or "user"
            count (int, optional): Number of rows to return. Defaults to 20.
            sort (str, optional): Order results by specific field and direction. Defaults to None.
                Accepts comma-seperated list of {field}:{direction} values.
                Supported fields are: created_at, Sort direction is specified as: asc, desc
            user (str, optional): If scope is set to "user", then user reference is required.

        Raises:
            CalendlyException: [description]

        Returns:
            dict: Json decoded response for Get Webhook request.
        """
        data = {'organization': organization,
                'scope': scope,
                'count': count}

        if (sort != None):
            data['sort'] = sort

        if (scope == 'user'):
            if (user == None):
                raise CalendlyException
            data['user'] = user

        response = self.request.get(WEBHOOK, data)
        return response.json()

    def delete_webhook(self, id: str) -> MutableMapping:
        """
        Delete a Webhook subscription for an Organization or User with a specified UUID.

        Args:
            id (str): Webhook UUID

        Returns:
            dict: Calenderly API response for delete webhook action.
        """
        dict_response = {'success': True}
        response = self.request.delete(f'{WEBHOOK}/{id}')
        dict_response['success'] = response.status_code == 200
        try:
            json_response = response.json()
        except json.JSONDecodeError:
            json_response = {}
        dict_response.update(json_response)
        return dict_response

    def get_webhook(self, uuid: str) -> MutableMapping:
        """
        Get a Webhook Subscription for an Organization or User with specified UUID.

        Args:
            uuid (str): Webhook uuid

        Returns:
            dict: Json decoded response from Calenderly API for Get webhook action.
        """
        response = self.request.get(f'{WEBHOOK}/{uuid}')
        return response.json()

    def about(self) -> MutableMapping:
        """
        Returns basic information about the user account.

        Returns:
            dict: Json decoded response about the basic information about the user.
        """
        response = self.request.get(ME)
        return response.json()

    def list_event_types(self, count: int=20, organization: str=None, page_token: str=None, sort: str=None, user_uri: str=None) -> MutableMapping:
        """
        Returns all Event Types associated with a specified user.

        Args:
            count (str, optional): Number of rows to return. Defaults to "20".
            organization (str, optional): View available personal, team and organization events type assosicated with the organization's URI. Defaults to None.
            page_token (str, optional): Toke to pass the next portion of the collection. Defaults to None.
            sort (str, optional): Order results by specified field and direction. Defaults to None.
            user_uri (str, optional): user's URI. Defaults to None.

        Returns:
            dict: json decoded response with list of event types
        """
        data = {"count": count}
        if organization:
            data['organization'] = organization
        if page_token:
            data['page_token'] = page_token
        if sort:
            data['sort'] = sort
        if user_uri:
            data['user'] = user_uri
        response = self.request.get(EVENT_TYPE, data)
        return response.json()

    def get_event_type(self, uuid: str) -> MutableMapping:
        """Returns event type associated with the specified UUID

        Args:
            uuid (str): Event UUID

        Returns:
            dict: json decoded response with information about the event
        """
        data = {"uuid": uuid}
        response = self.request.get(f'{EVENT_TYPE}/' + uuid, data)
        return response.json()

    def list_events(self, count: int=20, organization: str=None, sort: str=None, user_uri: str=None, status: str=None) -> MutableMapping:
        """
        Returns a List of Events

        Args:
            count (str, optional): Number of rows to return. Defaults to "20".
            organization (str, optional): Organization URI. Defaults to None.
            sort (str, optional): comma seperated list of {field}:{direction} values. Defaults to None.
            user_uri (str, optional): User URI. Defaults to None.
            status (str, optional): 'active' or 'canceled'. Defaults to None.

        Returns:
            dict: json decoded response of list of events.
        """
        data = {'count': count}
        if organization:
            data['organization'] = organization
        if sort:
            data['sort'] = sort
        if user_uri:
            data['user'] = user_uri
        if status:
            data['status'] = status
        response = self.request.get(EVENTS, data)
        return response.json()

    def get_event_invitee(self, event_uuid: str, invitee_uuid: str) -> MutableMapping:
        """
        Returns information about an invitee associated with a URI

        Args:
            event_uuid (str): Event's unique identifier
            invitee_uuid (str): Invitee's unique identifier

        Returns:
            dict: json decoded response about invitee information
        """
        url = f'{EVENTS}/' + event_uuid + '/invitees/' + invitee_uuid
        response = self.request.get(url)
        return response.json()

    def get_event_details(self, uuid: str) -> MutableMapping:
        """
        Get information about an Event associated with a URI.

        Args:
            uuid (str): Event's unique identifier

        Returns:
            dict: json decoded response
        """
        url = f'{EVENTS}/' + uuid
        response = self.request.get(url)
        return response.json()

    def list_event_invitees(self, uuid: str) -> List[MutableMapping]:
        """
        Returns a list of Invitees for an Event.

        Args:
            uuid (str): Event's unique identifier.

        Returns:
            dict: json decoded response
        """
        url = f'{EVENTS}/' + uuid + '/invitees'
        response = self.request.get(url)
        return response.json()

    def get_all_event_types(self, user_uri: str) -> List[MutableMapping]:
        """
        Get all event types by recursively crawling on all result pages.

        Args:
            user_uri (str, optional): User URI.

        Returns:
            list: json event type objects
        """
        first = self.list_event_types(user_uri=user_uri, count=100)
        next_page = first['pagination']['next_page']
        
        data = first['collection']

        while next_page:
            page = self.request.get(next_page).json()
            data += page['collection']
            next_page = page['pagination']['next_page']
        
        return data

    def get_all_scheduled_events(self, user_uri: str) -> List[MutableMapping]:
        """
        Get all scheduled events by recursively crawling on all result pages.

        Args:
            user_uri (str, optional): User URI.

        Returns:
            list: json scheduled event objects
        """
        first = self.list_events(user_uri=user_uri, count=100)
        next_page = first['pagination']['next_page']
        
        data = first['collection']

        while (next_page):
            page = self.request.get(next_page).json()
            data += page['collection']
            next_page = page['pagination']['next_page']
        
        return data

    def convert_event_to_original_url(self, event_uri: str, user_uri: str) -> str:
        """
        Convert event url from calendly's inner API convention to the original public url of the event.

        Args:
            event_uri (str): Event URI.
            user_uri (str, optional): User URI.

        Returns:
            string: the public convention of the event's url
        """
        event_type_uri = self.get_event_details(event_uri)['resource']['event_type']
        page = self.list_event_types(user_uri=user_uri)

        while True:
            filtered_result = next(filter(lambda event: event['uri'] == event_type_uri, page['collection']), None)
            if filtered_result:
                return filtered_result['scheduling_url']

            next_page = page['pagination']['next_page']
            if not next_page:
                break
            page = self.request.get(next_page).json()
        


from calendly.exceptions import CalendlyOauth2Exception, CalendlyException
from .api import CalendlyReq
from .constants import OAUTH_AUTHORIZE_URL, OAUTH_TOKEN_URL, OAUTH_REVOKE_URL, OAUTH_INTROSPECT_URL

__author__ = "luis <luiscastillocr@gmail.com>"
__license__ = "MIT"

REDIRECT_URI_EXCEPTION_TEXT = "You must pass the redirect_uri in the CalendlyOauth2 instantiation."

class CalendlyOauth2(object):
    """
    Private class wrapping the Calendly Oauth2 Api

    References
    ----------
    https://developer.calendly.com/how-to-authenticate-with-oauth

    """

    client_id = None
    client_secret = None
    redirect_uri = None
    response_type = None

    def __init__(self, client_id: str, client_secret: str, redirect_uri: str=None, response_type: str=None):
        """
        Constructor. client_id, client_secret, optionally you can pass the redirect_uri and response type

        Parameters
        ----------
        client_id : str
        client_secret : str
        redirect_uri : str
        response_type : str
        """
        self.request = CalendlyReq()
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.response_type = response_type or "code"

    @property
    def authorization_url(self):
        if not self.redirect_uri:
            raise CalendlyOauth2Exception(REDIRECT_URI_EXCEPTION_TEXT)

        return f"{OAUTH_AUTHORIZE_URL}?client_id={self.client_id}&response_type={self.response_type}&redirect_uri={self.redirect_uri}"

    def send_post(self, *args, **kwargs):
        try:
            return self.request.post(*args, **kwargs)
        except CalendlyException as e:
            raise CalendlyOauth2Exception(str(e))

    def get_access_token(self, code: str, grant_type: str=None):

        if not self.redirect_uri:
            raise CalendlyOauth2Exception(REDIRECT_URI_EXCEPTION_TEXT)

        data = dict(
            grant_type=grant_type or "authorization_code",
            client_id=self.client_id,
            client_secret=self.client_secret,
            code=code,
            redirect_uri=self.redirect_uri
        )

        response = self.send_post(OAUTH_TOKEN_URL, data)
        return response.json()

    def revoke_access_token(self, token: str):
        data = dict(
            client_id=self.client_id,
            client_secret=self.client_secret,
            token=token
        )

        response = self.send_post(OAUTH_REVOKE_URL, data)
        return response.json()

    def refresh_access_token(self, refresh_token: str, grant_type: str=None):
        data = dict(
            grant_type=grant_type or "refresh_token",
            client_id=self.client_id,
            client_secret=self.client_secret,
            refresh_token=refresh_token
        )

        response = self.send_post(OAUTH_TOKEN_URL, data)
        return response.json()

    def introspect_access_token(self, token: str):
        data = dict(
            client_id=self.client_id,
            client_secret=self.client_secret,
            token=token
        )

        response = self.send_post(OAUTH_INTROSPECT_URL, data)
        return response.json()





try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

long_desc = open("README.md").read()
required = ['requests']

setup(
    name='PyCalendly',
    version="0.1.0",
    author="laxmena",
    author_email="ConnectWith@laxmena.com",
    license="MIT",
    url="https://github.com/laxmena/PyCalendly",
    download_url="https://github.com/laxmena/PyCalendly",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    description="Python wrapper for Calendly API v2",
    packages=[
        'calendly',
        'calendly.utils'
    ],
    include_package_data=True,
    zip_safe=False,
    install_requires=required,
    platforms="any",
    keywords="Calendly python api v2",
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Programming Language :: Python"
    ],
    python_requires=">3.6",
)
